/*
**  $Id: driver4.c,v 1.1 1997/09/30 12:00:00 jochen Rel $
**  Copyright (C) 1997  Jochen Metzinger<1metzing@informatik.uni-hamburg.de>
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Library General Public
**  License as published by the Free Software Foundation; either
**  version 2 of the License, or (at your option) any later version.
**
**  This library is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
**  Library General Public License for more details.
**
**  You should have received a copy of the GNU Library General Public
**  License along with this library; if not, write to the Free
**  Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
**  MA 02111-1307, USA
*/

#include <string.h>
#include <vgagl.h>
#include <asm/io.h>
#include "driver4.h"


#define SEQ_I   0x3C4           /* Sequencer Index */
#define SEQ_D   0x3C5           /* Sequencer Data Register */

#define GRA_I   0x3CE           /* Graphics Controller Index */
#define GRA_D   0x3CF           /* Graphics Controller Data Register */


#define LowByte(w)  (*((unsigned char *)&(w) + 0))
#define HighByte(w) (*((unsigned char *)&(w) + 1))

 
#define ASSIGNVP4(x, y, vp) vp = VBUF + (((x) + (y)*WIDTH) >> 3);
#define ASSIGNMK4(x, y, mask) mask = 0x80 >> ((x) & 7);


static unsigned char saved_SEQ_mask;	/* 0x02 */

static unsigned char saved_GC_eSR;	/* 0x01 */
static unsigned char saved_GC_fun;	/* 0x03 */
static unsigned char saved_GC_rmap;	/* 0x04 */
static unsigned char saved_GC_mode;	/* 0x05 */
static unsigned char saved_GC_mask;	/* 0x08 */


static unsigned char leftMask;
static int byteCounter;
static unsigned char rightMask;
	/* results of get_masks() */


static void
get_masks(int x, int w)
{
	register int tmp;

	leftMask = rightMask = 0;
	byteCounter = w;
	/* right margin */
	tmp = (x+w) & 7;
	if (tmp) {
		byteCounter -= tmp;
		rightMask = (unsigned char)(0xff00 >> tmp);
	}
	/* left margin */
	tmp = x & 7;
	if (tmp) {
		byteCounter -= (8 - tmp);
		leftMask = (0xff >> tmp);
	}
	/* too small ? */
	if (byteCounter < 0) {
		leftMask &= rightMask;
		rightMask = 0;
		byteCounter = 0;
	}
	byteCounter /= 8;
}


void
driver4_setpixel(int x, int y, int c)
{
	unsigned char *vp;
	unsigned char mask;
	volatile unsigned char dummy;

	ASSIGNVP4(x, y, vp)
	ASSIGNMK4(x, y, mask)
	outb(0x05, GRA_I);	/* write mode 2 */
	saved_GC_mode = inb(GRA_D);
	outb(0x02, GRA_D);
	outb(0x03, GRA_I);	/* replace */
	saved_GC_fun = inb(GRA_D);
	outb(0x00, GRA_D);
	outb(0x08, GRA_I);	/* bit mask */
	saved_GC_mask = inb(GRA_D);
	outb(mask, GRA_D);
	dummy = *vp; *vp = c;
	/* reset GC register */
	outb(saved_GC_mask, GRA_D);
	outb(0x03, GRA_I);
	outb(saved_GC_fun, GRA_D);
	outb(0x05, GRA_I);
	outb(saved_GC_mode, GRA_D);
}


int
driver4_getpixel(int x, int y)
{
	volatile unsigned char *vp;
	unsigned char mask;
	unsigned char result;

	ASSIGNVP4(x, y, vp)
	ASSIGNMK4(x, y, mask)
	outb(0x05, GRA_I);	/* read mode 0 */
	saved_GC_mode = inb(GRA_D);
	outb(0x00, GRA_D);
	outb(0x04, GRA_I);	/* read map select */
	saved_GC_rmap = inb(GRA_D);
	result = 0x00;
	outb(0x00, GRA_D);	/* read map select */
	if (*vp & mask) result |= 0x01;
	outb(0x01, GRA_D);	/* read map select */
	if (*vp & mask) result |= 0x02;
	outb(0x02, GRA_D);	/* read map select */
	if (*vp & mask) result |= 0x04;
	outb(0x03, GRA_D);	/* read map select */
	if (*vp & mask) result |= 0x08;
	/* reset GC */
	outb(saved_GC_rmap, GRA_D);
	outb(0x05, GRA_I);
	outb(saved_GC_mode, GRA_D);
	return result;
}


void
driver4_hline(int x1, int y, int x2, int c)
{
	unsigned char *vp;
	volatile unsigned char dummy;

	ASSIGNVP4(x1, y, vp)
	get_masks(x1, x2-x1+1);
	outb(0x05, GRA_I);	/* write mode 2 */
	saved_GC_mode = inb(GRA_D);
	outb(0x02, GRA_D);
	outb(0x03, GRA_I);	/* replace */
	saved_GC_fun = inb(GRA_D);
	outb(0x00, GRA_D);
	outb(0x08, GRA_I);	/* bit mask */
	saved_GC_mask = inb(GRA_D);
	if (leftMask) {
		outb(leftMask, GRA_D);	/* bit mask */
		dummy = *vp; *vp++ = c;
	}
	if (byteCounter) {
		outb(0xff, GRA_D);	/* bit mask */
		memset(vp, c, byteCounter);
		vp += byteCounter;
	}
	if (rightMask) {
		outb(rightMask, GRA_D);	/* bit mask */
		dummy = *vp; *vp = c;
	}
	/* reset GC register */
	outb(saved_GC_mask, GRA_D);
	outb(0x03, GRA_I);
	outb(saved_GC_fun, GRA_D);
	outb(0x05, GRA_I);
	outb(saved_GC_mode, GRA_D);
}


void
driver4_fillbox(int x, int y, int w, int h, int c)
{
	unsigned char *vp, *vpX;
	volatile unsigned char dummy;
	int byte_per_line;
	int i;

	ASSIGNVP4(x, y, vpX)
	get_masks(x, w);
	byte_per_line = WIDTH >> 3;
	outb(0x05, GRA_I);	/* write mode 2 */
	saved_GC_mode = inb(GRA_D);
	outb(0x02, GRA_D);
	outb(0x03, GRA_I);	/* replace */
	saved_GC_fun = inb(GRA_D);
	outb(0x00, GRA_D);
	outb(0x08, GRA_I);	/* bit mask */
	saved_GC_mask = inb(GRA_D);
	if (leftMask) {
		outb(leftMask, GRA_D);	/* bit mask */
		/* write to video */
		vp = vpX;
		for (i=h; i>0; i--) {
			dummy = *vp; *vp = c;
			vp += byte_per_line;
		}
		vpX++;
	}
	if (byteCounter) {
		outb(0xff, GRA_D);	/* bit mask */
		/* write to video */
		vp = vpX;
		for (i=h; i>0; i--) {
			memset(vp, c, byteCounter);
			vp += byte_per_line;
		}
		vpX += byteCounter;
	}
	if (rightMask) {
		outb(rightMask, GRA_D);	/* bit mask */
		/* write to video */
		vp = vpX;
		for (i=h; i>0; i--) {
			dummy = *vp; *vp = c;
			vp += byte_per_line;
		}
	}
	/* reset GC register */
	outb(saved_GC_mask, GRA_D);
	outb(0x03, GRA_I);
	outb(saved_GC_fun, GRA_D);
	outb(0x05, GRA_I);
	outb(saved_GC_mode, GRA_D);
}


void
driver4_putbox(int x, int y, int w, int h, void *b, int bw)
{
	unsigned char *bp, *bpX;
	unsigned char *vp, *vpX;
	unsigned char mask;
	volatile unsigned char dummy;
	int byte_per_line;
	int i, j;

	bpX = b;
	ASSIGNVP4(x, y, vpX)
	ASSIGNMK4(x, y, mask)
	byte_per_line = WIDTH >> 3;
	outb(0x05, GRA_I);      /* write mode 2 */
	saved_GC_mode = inb(GRA_D);
	outb(0x02, GRA_D);
	outb(0x03, GRA_I);      /* replace */
	saved_GC_fun = inb(GRA_D);
	outb(0x00, GRA_D);
	outb(0x08, GRA_I);      /* bit mask */
	saved_GC_mask = inb(GRA_D);
	for (i=w; i>0; i--) {
		outb(mask, GRA_D);
		bp = bpX;
		vp = vpX;
		for (j=h; j>0; j--) {
			dummy = *vp; *vp = *bp;
			bp += bw;
			vp += byte_per_line;
		}
		bpX++;
		if ((mask >>= 1) == 0) {
			vpX++;
			mask = 0x80;
		}
	}
	/* reset GC register */
	outb(saved_GC_mask, GRA_D);
	outb(0x03, GRA_I);
	outb(saved_GC_fun, GRA_D);
	outb(0x05, GRA_I);
	outb(saved_GC_mode, GRA_D);
}


void
driver4_getbox(int x, int y, int w, int h, void *b, int bw)
{
	unsigned char *vp, *vpY, *vpP;
	unsigned char data, mask, maskP;
	unsigned char *bp, *bpY;
	unsigned char plane_mask;
	int byte_per_line = WIDTH >> 3;
	int plane, i, j;

	ASSIGNVP4(x, y, vpP)
	ASSIGNMK4(x, y, maskP)
	get_masks(x, w);
	outb(0x05, GRA_I);	/* read mode 0 */
	saved_GC_mode = inb(GRA_D);
	outb(0x00, GRA_D);
	outb(0x04, GRA_I);	/* read map select */
	saved_GC_rmap = inb(GRA_D);
	/* clear buffer */
	bp=b;
	for (j=h; j>0; j--) {
		memset(bp, 0, w);
		bp += bw;
	}
	for (plane=0, plane_mask=1; plane<4; plane++, plane_mask<<=1) {
		outb(plane, GRA_D);	/* read map select */
		vpY = vpP;
		bpY = b;
		for (j=h; j>0; j--) {
			vp = vpY;
			bp = bpY;
			if (leftMask) {
				mask = maskP;
				data = *vp++;
				do {
					if (data & mask) *bp |= plane_mask;
					bp++;
					mask >>= 1;
				} while (mask & leftMask);

			}
			if (byteCounter) {
				for (i=byteCounter; i>0; i--) {
					data = *vp++;
					if (data & 0x80) *bp |= plane_mask;
					bp++;
					if (data & 0x40) *bp |= plane_mask;
					bp++;
					if (data & 0x20) *bp |= plane_mask;
					bp++;
					if (data & 0x10) *bp |= plane_mask;
					bp++;
					if (data & 0x08) *bp |= plane_mask;
					bp++;
					if (data & 0x04) *bp |= plane_mask;
					bp++;
					if (data & 0x02) *bp |= plane_mask;
					bp++;
					if (data & 0x01) *bp |= plane_mask;
					bp++;
				}
			}
			if (rightMask) {
				mask = 0x80;
				data = *vp;
				do {
					if (data & mask) *bp |= plane_mask;
					bp++;
					mask >>= 1;
				} while (mask & rightMask);
			}
			bpY += bw;
			vpY += byte_per_line;
		}
	}
	/* reset GC register */
	outb(saved_GC_rmap, GRA_D);
	outb(0x05, GRA_I);
	outb(saved_GC_mode, GRA_D);
}
			
	
void
driver4_putboxmask(int x, int y, int w, int h, void *b)
{
	unsigned char *bp, *bpX;
	unsigned char *vp, *vpX;
	unsigned char color, mask;
	volatile unsigned char dummy;
	int byte_per_line;
	int i, j;

	bpX = b;
	ASSIGNVP4(x, y, vpX)
	ASSIGNMK4(x, y, mask)
	byte_per_line = WIDTH >> 3;
	outb(0x05, GRA_I);      /* write mode 2 */
	saved_GC_mode = inb(GRA_D);
	outb(0x02, GRA_D);
	outb(0x03, GRA_I);      /* replace */
	saved_GC_fun = inb(GRA_D);
	outb(0x00, GRA_D);
	outb(0x08, GRA_I);      /* bit mask */
	saved_GC_mask = inb(GRA_D);
	for (i=w; i>0; i--) {
		outb(mask, GRA_D);
		bp = bpX;
		vp = vpX;
		for (j=h; j>0; j--) {
			if ((color = *bp) != 0) {
				dummy = *vp; *vp = *bp;
			}
			bp += w;
			vp += byte_per_line;
		}
		bpX++;
		if ((mask >>= 1) == 0) {
			vpX++;
			mask = 0x80;
		}
	}
	/* reset GC register */
	outb(saved_GC_mask, GRA_D);
	outb(0x03, GRA_I);
	outb(saved_GC_fun, GRA_D);
	outb(0x05, GRA_I);
	outb(saved_GC_mode, GRA_D);
}


void
driver4_putboxpart(int x, int y, int w, int h, int ow, int oh,
			void *b, int xo, int yo)
{
	driver4_putbox(x, y, w, h, (char *)b + yo * ow + xo, ow);
}


void
driver4_getboxpart(int x, int y, int w, int h, int ow, int oh,
			void *b, int xo, int yo)
{
	driver4_getbox(x, y, w, h, (char *)b + yo * ow + xo, ow);
}



void
driver4_copybox(int x1, int y1, int w, int h, int x2, int y2)
{
	int direction;
	unsigned char *src, *srcZ;
	unsigned char *dst, *dstZ;
	unsigned char data;
	volatile unsigned char dummy;
	int i, j, plane;
	unsigned char plane_mask;
	unsigned short wData;
	int shiftA, shiftB, shiftOverflow;
	int byte_per_line;

	direction = ((y1 < y2) ? 0x010 : (y1 == y2) ? 0x020 : 0x030)
                  | ((x1 < x2) ? 0x101 : (x1 == x2) ? 0x202 : 0x303);
	if (direction == 0x222)
		return;
        get_masks(x2, w);
	byte_per_line = WIDTH >> 3;
	outb(0x05, GRA_I);	/* read mode 0, write mode 0 */
	saved_GC_mode = inb(GRA_D);
	outb(0x00, GRA_D);
	outb(0x04, GRA_I);      /* read map select */
	saved_GC_rmap = inb(GRA_D);
	outb(0x01, GRA_I);      /* use cpu wData */
	saved_GC_eSR = inb(GRA_D);
	outb(0x00, GRA_D);
	outb(0x03, GRA_I);      /* replace */
	saved_GC_fun = inb(GRA_D);
	outb(0x00, GRA_D);
	outb(0x08, GRA_I);      /* bit mask */
	saved_GC_mask = inb(GRA_D);
	outb(0x02, SEQ_I);      /* map mask */
	saved_SEQ_mask = inb(SEQ_D);
	shiftA = (x2 & 7) - (x1 & 7);
	if (shiftA && (direction & 0x0ff) > 0x022) {
		/* copy top-down,left-right with shift */
        	ASSIGNVP4(x1, y1, srcZ)
        	ASSIGNVP4(x2, y2, dstZ)
		shiftA = 8 + (x1 & 7) - (x2 & 7);
		shiftOverflow = (shiftA > 8);
		if (shiftOverflow) shiftA -= 8;
		shiftB = 8 - shiftA;
		for (j=h; j>0; j--) {
			plane_mask=1;
			for (plane=0; plane<4; plane++) {
				outb(0x04, GRA_I);	/* read map select */
				outb(plane, GRA_D);
				outb(0x02, SEQ_I);      /* map mask */
				outb(plane_mask, SEQ_D);
				src = srcZ;
				dst = dstZ;
				if (shiftOverflow)
					HighByte(wData) = *src++;
				outb(0x08, GRA_I);	/* bit mask */
				if (leftMask) {
					outb(leftMask, GRA_D);	/* bit mask */
					LowByte(wData) = *src++;
					wData <<= shiftA;
					dummy = *dst;
					*dst++ = HighByte(wData);
					wData <<= shiftB;
				}
				if (byteCounter) {
					outb(0xff, GRA_D);	/* bit mask */
					for (i=byteCounter; i>0; i--) {
						LowByte(wData) = *src++;
						wData <<= shiftA;
						*dst++ = HighByte(wData);
						wData <<= shiftB;
					}
				}
				if (rightMask) {
					outb(rightMask, GRA_D);	/* bit mask */
					LowByte(wData) = *src;
					wData <<= shiftA;
					dummy = *dst;
					*dst = HighByte(wData);
				}
				plane_mask <<= 1;
			}
			srcZ += byte_per_line;
			dstZ += byte_per_line;
		}
	} else if (shiftA) {
		/* copy bottom-up,right-left with shift */
        	ASSIGNVP4(x1+=w-1, y1+=h-1, srcZ)
        	ASSIGNVP4(x2+=w-1, y2+=h-1, dstZ)
		shiftA = 8 + (x2 & 7) - (x1 & 7);
		shiftOverflow = (shiftA > 8);
		if (shiftOverflow) shiftA -= 8;
		shiftB = 8 - shiftA;
		for (j=h; j>0; j--) {
			plane_mask=1;
			for (plane=0; plane<4; plane++) {
				outb(0x04, GRA_I);	/* read map select */
				outb(plane, GRA_D);
				outb(0x02, SEQ_I);      /* map mask */
				outb(plane_mask, SEQ_D);
				src = srcZ;
				dst = dstZ;
				if (shiftOverflow)
					LowByte(wData) = *src--;
				outb(0x08, GRA_I);	/* bit mask */
				if (rightMask) {
					outb(rightMask, GRA_D);
					HighByte(wData) = *src--;
					wData >>= shiftA;
					dummy = *dst;
					*dst-- = LowByte(wData);
					wData >>= shiftB;
				}
				if (byteCounter) {
					outb(0xff, GRA_D);	/* bit mask */
					for (i=byteCounter; i>0; i--) {
						HighByte(wData) = *src--;
						wData >>= shiftA;
						*dst-- = LowByte(wData);
						wData >>= shiftB;
					}
				}
				if (leftMask) {
					outb(leftMask, GRA_D);	/* bit mask */
					HighByte(wData) = *src;
					wData >>= shiftA;
					dummy = *dst;
					*dst = LowByte(wData);
				}
				plane_mask <<= 1;
			}
			srcZ -= byte_per_line;
			dstZ -= byte_per_line;
		}
	} else if (direction > 0x22F) {
		/* copy left-right,top-down w/o shift */
        	ASSIGNVP4(x1, y1, srcZ)
        	ASSIGNVP4(x2, y2, dstZ)
		if (leftMask) {
			outb(0x08, GRA_I);	/* bit mask */
			outb(leftMask, GRA_D);
			plane_mask=1;
			for (plane=0; plane<4; plane++) {
				outb(0x04, GRA_I);	/* read map select */
				outb(plane, GRA_D);
				outb(0x02, SEQ_I);      /* map mask */
				outb(plane_mask, SEQ_D);
				src = srcZ;
				dst = dstZ;
				for (j=h; j>0; j--) {
					data = *src;
					dummy = *dst; *dst = data;
					src += byte_per_line;
					dst += byte_per_line;
				}
				plane_mask <<= 1;
			}
			srcZ++;
			dstZ++;
		}
		if (byteCounter) {
			outb(0x02, SEQ_I);      /* map mask */
			outb(0x0f, SEQ_D);
			outb(0x05, GRA_I);	/* read mode 0, write mode 1 */
			outb(0x01, GRA_D);
			for (i=byteCounter; i>0; i--) {
				src = srcZ++;
				dst = dstZ++;
				for (j=h; j>0; j--) {
					*dst = *src;
					src += byte_per_line;
					dst += byte_per_line;
				}
			}
			outb(0x05, GRA_I);	/* read mode 0, write mode 0 */
			outb(0x00, GRA_D);
		}
		if (rightMask) {
			outb(0x08, GRA_I);	/* bit mask */
			outb(rightMask, GRA_D);
			plane_mask=1;
			for (plane=0; plane<4; plane++) {
				outb(0x04, GRA_I);	/* read map select */
				outb(plane, GRA_D);
				outb(0x02, SEQ_I);      /* map mask */
				outb(plane_mask, SEQ_D);
				src = srcZ;
				dst = dstZ;
				for (j=h; j>0; j--) {
					data = *src;
					dummy = *dst; *dst = data;
					src += byte_per_line;
					dst += byte_per_line;
				}
				plane_mask <<= 1;
			}
		}
	} else {
		/* copy right-left,bottom-up w/o shift */
        	ASSIGNVP4(x1+=w-1, y1+=h-1, srcZ)
        	ASSIGNVP4(x2+=w-1, y2+=h-1, dstZ)
		if (rightMask) {
			outb(0x08, GRA_I);	/* bit mask */
			outb(rightMask, GRA_D);
			plane_mask=1;
			for (plane=0; plane<4; plane++) {
				outb(0x04, GRA_I);	/* read map select */
				outb(plane, GRA_D);
				outb(0x02, SEQ_I);      /* map mask */
				outb(plane_mask, SEQ_D);
				src = srcZ;
				dst = dstZ;
				for (j=h; j>0; j--) {
					data = *src;
					dummy = *dst; *dst = data;
					src -= byte_per_line;
					dst -= byte_per_line;
				}
				plane_mask <<= 1;
			}
			srcZ--;
			dstZ--;
		}
		if (byteCounter) {
			outb(0x02, SEQ_I);      /* map mask */
			outb(0x0f, SEQ_D);
			outb(0x05, GRA_I);	/* read mode 0, write mode 1 */
			outb(0x01, GRA_D);
			for (i=byteCounter; i>0; i--) {
				src = srcZ--;
				dst = dstZ--;
				for (j=h; j>0; j--) {
					*dst = *src;
					src -= byte_per_line;
					dst -= byte_per_line;
				}
			}
			outb(0x05, GRA_I);	/* read mode 0, write mode 0 */
			outb(0x00, GRA_D);
		}
		if (leftMask) {
			outb(0x08, GRA_I);	/* bit mask */
			outb(leftMask, GRA_D);
			plane_mask=1;
			for (plane=0; plane<4; plane++) {
				outb(0x04, GRA_I);	/* read map select */
				outb(plane, GRA_D);
				outb(0x02, SEQ_I);      /* map mask */
				outb(plane_mask, SEQ_D);
				src = srcZ;
				dst = dstZ;
				for (j=h; j>0; j--) {
					data = *src;
					dummy = *dst; *dst = data;
					src -= byte_per_line;
					dst -= byte_per_line;
				}
				plane_mask <<= 1;
			}
		}
	}
	/* reset SEQ */
	outb(0x02, SEQ_I);
	outb(saved_SEQ_mask, SEQ_D);
	/* reset GC */
	outb(0x08, GRA_I);
	outb(saved_GC_mask, GRA_D);
	outb(0x03, GRA_I);
	outb(saved_GC_fun, GRA_D);
	outb(0x01, GRA_I);      /* use cpu wData */
	outb(saved_GC_eSR, GRA_D);
	outb(0x04, GRA_I);
	outb(saved_GC_rmap, GRA_D);
	outb(0x05, GRA_I);
	outb(saved_GC_mode, GRA_D);
}
